/*
Copyright 2016-17 Intel Corporation.

The source code, information and material ("Material") contained herein
is owned by Intel Corporation or its suppliers or licensors, and title
to such Material remains with Intel Corporation or its suppliers or
licensors. The Material contains proprietary information of Intel or
its suppliers and licensors. The Material is protected by worldwide
copyright laws and treaty provisions. No part of the Material may be used,
copied, reproduced, modified, published, uploaded, posted, transmitted,
distributed or disclosed in any way without Intel's prior express written
permission. No license under any patent, copyright or other intellectual
property rights in the Material is granted to or conferred upon you,
either expressly, by implication, inducement, estoppel or otherwise. Any
license under such intellectual property rights must be express and
approved by Intel in writing.

Unless otherwise agreed by Intel in writing, you may not remove or alter
this notice or any other notice embedded in Materials by Intel or Intel's
suppliers or licensors in any way.
*/

#ifdef WIN32
#include <Windows.h>
#include <io.h>
#include <stdio.h>
#include <tchar.h>
#else
#include <dlfcn.h>
#include <sgx_capable.h>
#endif

#include <iostream>
#include <fcntl.h>
#include <sgx_uae_service.h>

using namespace std;

#define ST_SGX_UNSUPPORTED				0x0 // SGX is unsupported
#define ST_SGX_CAPABLE					0x1 // The CPU is SGX-capable, but not necessary enabled
#define ST_SGX_ENABLED					0x2	// SGX instructions are supported
#define ST_SGX_REBOOT_REQUIRED			0x4 // SGX will be enabled after reboot
#define ST_SGX_BIOS_ENABLE_REQUIRED		0x8 // SGX must be explicitly enabled via the BIOS

int platform_supports_sgx(unsigned int *sgx_support);

#ifdef WIN32
int is_psw_installed (HINSTANCE *h_urts, HINSTANCE *h_service);
#else
int is_psw_installed (void **h_urts, void **h_service);
typedef sgx_status_t (*sgx_cap_get_status_t)(sgx_device_status_t *);
#endif
typedef sgx_status_t (*sgx_enable_device_t)(sgx_device_status_t *);

#ifdef WIN32
int _tmain(int argc, _TCHAR* argv[])
#else 
int main(int argc, char* argv[])
#endif
{
	unsigned int sgx_support;
	int rv= 0;

#ifdef WIN32
	_setmode(_fileno(stdout), _O_U16TEXT);
	_setmode(_fileno(stderr), _O_U16TEXT);
#endif

	if ( platform_supports_sgx(&sgx_support) == 0 ) {
		wcerr << L"Could not determine status of Intel(R) SGX support on this system."  << endl;
	} else {
		if ( sgx_support & ST_SGX_CAPABLE ) {
			rv= 1;

			if ( sgx_support & ST_SGX_REBOOT_REQUIRED ) {
				wcout << L"Intel(R) SGX will be available after you reboot your system."  << endl;
			} else if ( sgx_support & ST_SGX_BIOS_ENABLE_REQUIRED ) {
				wcout << L"You must enable Intel(R) SGX in your BIOS before you can use SGX applications."  << endl;
			} else if ( sgx_support & ST_SGX_ENABLED ) {
				wcout << L"Intel(R) SGX is supported on this system."  << endl;
			} else {
				wcout << L"This system is capable of Intel(R) SGX but it's not available in the current environment."  << endl;
			}
		} else {
			wcout << L"This system does not support Intel(R) SGX."  << endl;
		}
	}

#ifdef WIN32
	wcout << "Press ENTER to exit..."  << endl;
	(void) getchar();
#endif

	return rv;
}

int platform_supports_sgx (unsigned int *sgx_support)
{
	sgx_status_t status;
	int rv= 0;
	int psw_status;
	sgx_device_status_t sgx_device_status;
#ifdef WIN32
	HINSTANCE h_urts, h_service;
#else
	void *h_urts, *h_service;
#endif
	sgx_enable_device_t p_sgx_enable_device= NULL;

	*sgx_support= ST_SGX_UNSUPPORTED;

	// Check for the PSW

	if ( (psw_status= is_psw_installed(&h_urts, &h_service)) != 1 ) {
		if ( psw_status == -1 ) {
			wcout << L"Can't determine whether or not the Platform Software is installed. Aborting."  << endl;
			rv= 0;
		} else if ( psw_status == 0 ) {
			// Need to install the PSW
			wcout << L"Platform Software not installed."  << endl;
			rv= 1;
		}

#ifdef WIN32
		if ( h_urts != NULL ) FreeLibrary(h_urts);
		if ( h_service != NULL ) FreeLibrary(h_service);
#else
		if ( h_urts != NULL ) dlclose(h_urts);
		if ( h_service != NULL ) dlclose(h_service);
#endif

		return rv;
	}

	*sgx_support= ST_SGX_CAPABLE;

	// Check to see if we have sgx_enable_device. The Windows PSW supports
	// this, but Linux versions up through 1.9 do not, and future versions
	// may or may not. By checking, we make ourselves future-proof, and
	// fall back to sgx_cap_get_status() if it's not there.

#ifdef WIN32
	p_sgx_enable_device= (sgx_enable_device_t) GetProcAddress(h_service, "sgx_enable_device");

	if ( p_sgx_enable_device == NULL ) {
		if ( h_urts != NULL ) FreeLibrary(h_urts);
		if ( h_service != NULL ) FreeLibrary(h_service);
		return rv;
	}

	status=  p_sgx_enable_device(&sgx_device_status);
#else
	p_sgx_enable_device= (sgx_enable_device_t) dlsym(h_service, "sgx_enable_device");
	if ( p_sgx_enable_device != NULL ) {
		status= p_sgx_enable_device(&sgx_device_status);
	} else {
		// Fall back to sgx_cap_get_status

		status= sgx_cap_get_status(&sgx_device_status);
	}
#endif

	// Free up these libraries 

#ifdef WIN32
	if ( h_urts != NULL ) FreeLibrary(h_urts);
	if ( h_service != NULL ) FreeLibrary(h_service);
#else
	if ( h_urts != NULL ) dlclose(h_urts);
	if ( h_service != NULL ) dlclose(h_service);
#endif

	if ( status != SGX_SUCCESS ) {
		switch(status) {
#ifdef WIN32
		case SGX_ERROR_SERVICE_UNAVAILABLE:
			wcout << L"The AE service did not respond. Make sure the 'Intel(R) SGX AESM' service is running."  << endl;
			return 0;
		case SGX_ERROR_HYPERV_ENABLED:
			wcout << L"This version of Windows 10 is incompatible with Hyper-V. Disable Hyper-V and try again."  << endl;
			return 1;
#endif
#ifdef SGX_ERROR_VMM_INCOMPATIBLE
		case SGX_ERROR_VMM_INCOMPATIBLE:
			wcout << L"The virtual machine monitor is incompatible with Intel(R) SGX."  << endl;
			return 1;
#endif
		case SGX_ERROR_INVALID_PARAMETER:
		case SGX_ERROR_UNEXPECTED:
			wcout << L"An unexpected error occurred."  << endl;
		}
		return 0;
	}
	
	// Check the SGX device status.
	
	if ( sgx_device_status != SGX_ENABLED ) {
		switch (sgx_device_status) {
		case SGX_DISABLED_REBOOT_REQUIRED:
			// Notify the user that they will need to reboot their system.
			*sgx_support|= ST_SGX_REBOOT_REQUIRED;
			break;
		case SGX_DISABLED_LEGACY_OS:
			*sgx_support|= ST_SGX_BIOS_ENABLE_REQUIRED;
			break;
		case SGX_DISABLED:
			// For completeness. We wouldn't actually get this far if SGX were disabled.
			break;
		}

		return 1;
	}

	*sgx_support|= ST_SGX_ENABLED;

	return 1;
}

// Is the PSW (Platform Software) installed?

#ifdef WIN32
int is_psw_installed (HINSTANCE *h_urts, HINSTANCE *h_service)
{

	_TCHAR *systemdir;
	unsigned int rv, sz;
	*h_urts= *h_service= NULL;

	// Get the system directory path. Start by finding out how much space we need
	// to hold it.

	sz= GetSystemDirectory(NULL, 0);
	if ( sz == 0 ) {
		wcerr << L"Could not determine Windows system directory." << endl;
		return -1;
	}

	systemdir= new _TCHAR[sz+1];
	rv= GetSystemDirectory(systemdir, sz);
	if ( rv == 0 || rv > sz ) {
		wcerr << L"Could not determine Windows system directory." << endl;
		return -1;
	}

	// Set our DLL search path to just the System directory so we don't accidentally
	// load the DLLs from an untrusted path.

	if ( SetDllDirectory(systemdir) == 0 ) {
		cerr << "Could not set DLL search path to " << systemdir << endl;
		delete systemdir;
		return -1;
	}

	delete systemdir; // No longer need this

	// Need to be able to load both of these DLLs from the System directory.

	if ( (*h_service= LoadLibrary(_T("sgx_uae_service.dll"))) == NULL ) {
		return 0;
	}

	if ( (*h_urts= LoadLibrary(_T("sgx_urts.dll"))) == NULL ) {
		FreeLibrary(*h_service);
		*h_service= NULL;
		return 0;
	} 

	return 1;
}
#else
int is_psw_installed (void **h_urts, void **h_service)
{
	*h_service= dlopen("libsgx_uae_service.so", RTLD_NOW);
	if ( h_service == NULL ) return 0;

	*h_urts= dlopen("libsgx_urts.so", RTLD_NOW);
	if ( h_urts == NULL ) {
		dlclose(*h_service);
		return 0;
	}

	return 1;
}
#endif
