/*
Copyright 2016-17 Intel Corporation.

The source code, information and material ("Material") contained herein
is owned by Intel Corporation or its suppliers or licensors, and title
to such Material remains with Intel Corporation or its suppliers or
licensors. The Material contains proprietary information of Intel or
its suppliers and licensors. The Material is protected by worldwide
copyright laws and treaty provisions. No part of the Material may be used,
copied, reproduced, modified, published, uploaded, posted, transmitted,
distributed or disclosed in any way without Intel's prior express written
permission. No license under any patent, copyright or other intellectual
property rights in the Material is granted to or conferred upon you,
either expressly, by implication, inducement, estoppel or otherwise. Any
license under such intellectual property rights must be express and
approved by Intel in writing.

Unless otherwise agreed by Intel in writing, you may not remove or alter
this notice or any other notice embedded in Materials by Intel or Intel's
suppliers or licensors in any way.
*/

#include "installer.h"
#ifdef WIN32
#include <tchar.h>
#include <stdio.h>
#include <Windows.h>
#include <winerror.h>
#include <io.h>
#else
#include <dlfcn.h>
#include <stdlib.h>
#endif
#include <sgx_capable.h>
#include <iostream>
#include <stdio.h>
#include <fcntl.h>

using namespace std;

int platform_supports_sgx (unsigned int *sgx_support);
int install_psw (int *status);

#define ST_SGX_UNSUPPORTED				0x0 // SGX is unsupported
#define ST_SGX_CAPABLE					0x1 // The CPU is SGX-capable, but not necessary enabled
#define ST_SGX_ENABLED					0x2	// SGX instructions are supported
#define ST_SGX_REBOOT_REQUIRED			0x4 // SGX will be enabled after reboot
#define ST_SGX_BIOS_ENABLE_REQUIRED		0x8 // SGX must be explicitly enabled via the BIOS

#ifdef WIN32
int _tmain(int argc, _TCHAR* argv[])
#else
int main(int argc, char* argv[])
#endif
{
	unsigned int sgx_support;
	int rv= 0;

#ifdef WIN32
	_setmode(_fileno(stdout), _O_U16TEXT);
	_setmode(_fileno(stderr), _O_U16TEXT);
#endif

	if ( platform_supports_sgx(&sgx_support) == 0 ) {
		wcerr << L"Could not determine status of Intel(R) SGX support on this system."  << endl;
	} else {
		if ( sgx_support & ST_SGX_CAPABLE ) {
			rv= 1;

			if ( sgx_support & ST_SGX_REBOOT_REQUIRED ) {
				wcerr << L"Intel(R) SGX will be available after you reboot your system."  << endl;
			} else if ( sgx_support & ST_SGX_BIOS_ENABLE_REQUIRED ) {
				wcerr << L"You must enable Intel(R) SGX in your BIOS before you can use SGX applications."  << endl;
			} else if ( sgx_support & ST_SGX_ENABLED ) {
				wcout << L"Intel(R) SGX is supported on this system."  << endl;
			} else {
				wcout << L"This system is capable of Intel(R) SGX but it's not available in the current environment."  << endl;
			}
		} else {
			wcout << L"This system does not support Intel(R) SGX."  << endl;
		}
	}

#ifdef WIN32
	wcout << L"Press ENTER to exit..."  << endl;
	(void) getchar();
#endif

	return rv;
}

int platform_supports_sgx (unsigned int *sgx_support)
{
	sgx_status_t rv;
	sgx_device_status_t sgx_device_status;
	int sgx_capable;
	int psw_status;
	
	*sgx_support= ST_SGX_UNSUPPORTED;

	// Are we capable of executing SGX instructions?
	
	if ( (rv= sgx_is_capable(&sgx_capable)) != SGX_SUCCESS ) {
		switch (rv) {
#ifdef SGX_ERROR_EFI_NOT_SUPPORTED
		case SGX_ERROR_EFI_NOT_SUPPORTED:
			wcout << L"Cannot query EFI for Intel(R) SGX support."  << endl;
			break;
#endif
		case SGX_ERROR_NO_PRIVILEGE:
			wcout << L"This program requires Administrator privileges."  << endl;
			break;
		case SGX_ERROR_UNEXPECTED:
		case SGX_ERROR_INVALID_PARAMETER:
			wcout << L"An unexpected error occurred."  << endl;
		}

		return 0;
	}

	if ( sgx_capable == 0 ) {
		wcout << L"This system does not support Intel(R) SGX." << endl;
		return 1;
	}

	*sgx_support = ST_SGX_CAPABLE;

	// Install/update the PSW

	if ( install_psw(&psw_status) == 0 ) {
#ifdef WIN32
		wchar_t err[1024];
		
		if (FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_ARGUMENT_ARRAY, NULL, GetLastError(), 0, err, 1023, NULL)) {\
			wcerr << L"Could not run Platform Software installer. CreateProcess: " << err << ""  << endl;
		}
#else
		wcerr << L"Could not run Platform Software installer." << endl;
#endif
		psw_status= 0;
	}

	if ( psw_status == 0 ) {
		wcout << L"Platform Software installation failed. Aborting."  << endl;
		return 1;
	}

	// Try to enable SGX

	if ( (rv= sgx_cap_enable_device(&sgx_device_status)) != SGX_SUCCESS ) {
		switch(rv) {
		case SGX_ERROR_NO_PRIVILEGE:
			wcout << L"This program requires Administrator privileges.";
			return 0;
#ifdef SGX_ERROR_EFI_NOT_SUPPORTED
		case SGX_ERROR_EFI_NOT_SUPPORTED:
			wcout << L"This system does not support the EFI interface. Cannot enable SGX via software.";	
			*sgx_support|= ST_SGX_BIOS_ENABLE_REQUIRED;
			return 1;
#endif
#ifdef WIN32
		case SGX_ERROR_HYPERV_ENABLED:
			wcout << L"This version of Windows 10 is incompatible with Hyper-V. Disable Hyper-V and try again.";
			return 1;
#endif
#ifdef SGX_ERROR_VMM_INCOMPATIBLE
		case SGX_ERROR_VMM_INCOMPATIBLE:
			wcout << L"The virtual machine monitor is incompatible with Intel(R)SGX.";
			return 1;
#endif
		case SGX_ERROR_INVALID_PARAMETER:
		case SGX_ERROR_UNEXPECTED:
			wcout << L"An unexpected error occurred.";
		}
		return 1;
	}
	
	// If SGX isn't enabled yet, perform the software opt-in/enable.
	
	if ( sgx_device_status != SGX_ENABLED ) {
		switch (sgx_device_status) {
		case SGX_DISABLED_REBOOT_REQUIRED:
			// Notify the user that they will need to reboot their system.
			*sgx_support|= ST_SGX_REBOOT_REQUIRED;
			break;
		case SGX_DISABLED_LEGACY_OS:
#ifdef SGX_DISABLED_MANUAL_ENABLE
		case SGX_DISABLED_MANUAL_ENABLE:
#endif
			*sgx_support|= ST_SGX_BIOS_ENABLE_REQUIRED;
			break;
		// These cases are for completeness. We wouldn't get past sgx_is_capable() if these
		// conditions were true.
#ifdef SGX_DISABLED_UNSUPPORTED_CPU
		case SGX_DISABLED_UNSUPPORTED_CPU:
			return 0;
			break;
#endif
		case SGX_DISABLED:
			break;
		}

		return 1;
	}

	*sgx_support|= ST_SGX_ENABLED;

	return 1;
}

int install_psw (int *status) {
	// Launch the installer. The PSW installer must be bundled with 
	// your application installer.
	//
	// We have hardcoded particular PSW install packages in install.h
	// for demonstration purposes.
#ifdef TEST_INSTALLER
# ifdef WIN32
	// It assumes the installer is in the same directory as the executable 
	// (this probably won't run as-is from within Visual Studio: you will 
	// need to put the installer in the same folder as the EXE and run it
	// from that directory).
	//
	// If you want more control over the installer, extract the MSI from
	// the EXE and use MISEXEC to do the installation.
	
	TCHAR cmdline[2048]; // Big enough for our purposes
	PROCESS_INFORMATION pi;
	STARTUPINFO si;
	DWORD exitcode;

	*status= 0;
	
	GetSystemDirectory(cmdline, 1024); // Plenty of room left for our install cmd and args

	lstrcat(cmdline, L"Intel(R)_SGX_Windows_x64_PSW_1.7.100.35600.exe");
	
	// Initialize si to 0 to get defaults.
	// Must set si.cb to size of the structure.
	
	ZeroMemory(&si, sizeof(si));
	si.cb= sizeof(si);

	// Is this necessary?
	ZeroMemory(&pi, sizeof(pi));
	
	if ( ! CreateProcess(
		NULL,
		cmdline, 
		NULL, NULL, 
		FALSE, 0,
		NULL, NULL, 
		&si, &pi
	) )	return 0;

	// Wait for the installer to exit.

	WaitForSingleObject(pi.hProcess, INFINITE);

	// We won't get here until the process exits.

	GetExitCodeProcess(pi.hProcess, &exitcode);

	// Check the exit code of MSIEXEC.
	// https://msdn.microsoft.com/en-us/library/windows/desktop/aa368542(v=vs.85).aspx

	if ( exitcode == ERROR_SUCCESS ) *status= 1;

# else
	// Execute the installer. Force the execution from the current
	// directory.

	int exitcode;

	/*----------------------------------------------------------------------
	 * Uncomment this to test the installer.
	exitcode= system("./"INSTALL_PKG);
	if ( exitcode == -1 || exitcode >= 127 ) return 0;

	if ( exitcode == 0 ) *status= 1;
	*----------------------------------------------------------------------*/

# endif
#else
	wcout << L">>> Skipping test of the PSW installer." << endl;
	wcout << L">>> To test installer, #define TEST_INSTALLER in install.h and recompile." << endl;
	*status= 1;
#endif
	return 1;
}
